/**********************************************************************
 * File Name: MobileObject.java
 *
 * Copyright (C) 2000, US Air Force Research Laboratory,
 * Rome Research Site, Rome NY 13424, USA.
 *
 * Use of this software implies acceptance of the terms described in
 * accompanying license.txt file.
 *
 *
 * Authors: Jennifer Dzek, James Lawton, John Crowter,
 *          Bob Paragi, and Joe Beasock
 *
 * Creation Date: July 1999
 *
 * Description:
 *
 * Modification History:
 *
 *  - June 00, Bryan Horling, UMass (BCH): Changes to make drawing more 
 *  efficient by only redrawing when needed.
 *
 **********************************************************************/

import java.lang.*;
import java.math.*;
import java.awt.*;
import java.util.*;
import java.io.*;

public class MobileObject extends SimObject {
    // Hack to estimate current screen size (BCH)
    protected int lastw, lasth;

    protected double x_coord;
    protected double y_coord;
    public double speed;
    public double mdirection;
    protected int last_update;
    public int mobile_object_tick;
    public boolean scanned;
    public boolean isMoving;

    // Event Types:
    public final int TICK_EVENT=1;


    public MobileObject() {
	name = new String("Unknown");
    }

    public MobileObject (String args, double width, double height) throws Exception {
	StringTokenizer s = new StringTokenizer(args);
	int n = s.countTokens();
	boolean bad_data = false;

	if (n == 6) {
	    s.nextToken();               // Get rid of the "MOBOBJ" tag
	    name = s.nextToken();
	    x_coord = (new Double(s.nextToken())).doubleValue();
	    if ((x_coord < 0.0) || (x_coord > width))
		bad_data = true;
	    y_coord = (new Double(s.nextToken())).doubleValue();
	    if ((y_coord < 0.0) || (y_coord > height))
		bad_data = true;
	    speed = (new Double(s.nextToken())).doubleValue();
	    if (speed < 0.0)
		bad_data = true;
	    mdirection = (new Double(s.nextToken())).doubleValue();
	    if ((mdirection < 0.0) || (mdirection > 360.0))
		bad_data = true;
	    scanned = false;
	    isMoving = false;
	}
	else throw (new Exception("Wrong number of args"));
	if (bad_data) throw (new Exception("argument(s) out of range"));
    }

    public MobileObject (String newname, double xptr, double yptr, double spd, double dir) {
	name = newname;
	x_coord = xptr;
	y_coord = yptr;
	speed = spd;
	mdirection = dir;
	last_update = 0;
	mobile_object_tick = 1;
	isMoving = false;
    }


    // After creation initialization:
    public void initMobileObject(int start_time) {
	// Log initial information
	Log.defTarget(name, env.clock.simtime, x_coord, y_coord, mdirection, speed);

	// Schedule initial event (i.e. when to start moving)
	schedule(TICK_EVENT, start_time);
	last_update = start_time;
    }

    // Shut down the object cleanly
    public void shutdown() {
	// Final logging info
	Log.targetState(name, env.clock.simtime, x_coord, y_coord, mdirection, speed);
    }

    public void drawSelf (Graphics g, int screen_width, int screen_height) {
	Color redColor = Color.red;
	Color whiteColor = Color.white;
	Color blackColor = Color.black;
	int x = env.scale_x_coord ( x_coord, (double) screen_width),
	    y = env.scale_y_coord ( y_coord, (double) screen_height),
	    rad = 10,
	    srad = rad + 4;
	lastw = screen_width; // BCH
	lasth = screen_height;

	// System.out.println (name + "X= " + (int) x_coord + "Y= " + (int) y_coord);
	if (scanned) {
	    g.setColor(whiteColor);
	    g.fillOval ( x - srad, y - srad, 2 * srad, 2 * srad);
	}
	g.setColor (redColor);
	g.fillOval ( x - rad, y - rad, 2 * rad, 2 * rad);
	g.setColor (whiteColor);

	// BCH
	FontMetrics fm = g.getFontMetrics();
	g.drawString(name, x - fm.stringWidth(name)/2, y + fm.getAscent()/2);
	// g.drawString (name, x - 10, y + 5);
	g.setColor (blackColor);
	g.drawString ("(" + (int) x_coord + "," + (int) y_coord + ")", x + 10, y);
    }

    public void dispatchEvent (EventStruct ev) {
	int etype = ev.event_type;
	int simtime = ev.event_time;

	// Reset this for the graphics
	scanned = false;

	if (env.Debug > 1)
	    System.out.println("MobileObject " + name + " dispatching event type: " + etype + " time: " + simtime);

	switch (etype) {
	case TICK_EVENT:
	    do_tick (simtime);
	    break;
	}
    }


    //-------------------------------------------------------
    // do_tick()
    //   Handle the SENSOR_TICK event
    //
    private void do_tick(int simtime) {
	isMoving = true;
        next_loc(simtime);
        Log.targetState(name, simtime, x_coord, y_coord, mdirection, speed);
        schedule(TICK_EVENT, simtime+1);
    }

    public boolean between (double x, double y, double z) {
	if ((x <= y) && (y <= z))
	    return true;
	else if ((x >= y) && (y >= z))
	    return true;
	else return false;
    }

    // if x = n where does line intersect it?
    public double intersect_x_at(double x, double y, double a, double b, double x_eq) {
	double slope = (y - b)/(x - a);
	double y_intercept = y - (slope * x);
	return ((slope * x_eq) + y_intercept);
    }

    // if y = n, where does line intersect it?
    public double intersect_y_at(double x, double y, double a, double b, double y_eq) {
	double slope = (y - b)/(x - a);
	double y_intercept = y - (slope * x);
	return (y_eq - y_intercept) / slope; // find x
    }
    public boolean outside_boundaries(double x, double y) {
	if (x > env.width) return true;
	else if (x < 0) return true;
	else if (y > env.height) return true;
	else if (y < 0) return true;
	else return false;
    }


    // Update our location base on how long its been since the
    // last location update.
    public void next_loc (int simtime) {
	double time, radian, sx, sy,nx, ny, tx, ty;

	int ox = env.scale_x_coord ( x_coord, (double) lastw), 
	    oy = env.scale_y_coord ( y_coord, (double) lasth); //BCH

	// Compute time in seconds
	time = (simtime - last_update) * ((double) env.clock.ticksize / 1000.0);
	last_update = simtime;

	if (time > 0) {
	    radian = (mdirection) * (Math.PI/180);
	    sx = x_coord;
	    sy = y_coord;

	    nx = x_coord + (time * speed * Math.cos(radian));
	    ny = y_coord + (time * speed * Math.sin(radian));
	    while (outside_boundaries(nx,ny)) {
		if (sx == nx){
		    if (sy < ny){
			sy = env.height;
			ny = sy - (ny - sy);
		    }
		    else {
			sy = 0.0;
			ny = sy - ny;
		    }
		}
		else if (sy == ny) {
		    if (sx < nx){
			sx = env.width;
			nx = sx - (nx - sx);
		    }
		    else {
			sx = 0.0;
			nx = sx - nx;
		    }
		}
		else if ((sx < nx) && (sy < ny)) { // first quadrant
		    ty = intersect_x_at(sx,sy,nx,ny,env.width);
		    tx = intersect_y_at(sx,sy,nx,ny,env.height);
		    if ((tx == env.width) && (ty == env.height)) {
			// intersects exactly in the corner (100,100)
			tx = env.height - (ny - env.height);
			ty = env.width - (nx - env.width);
			sx = env.width;
			sy = env.height;
			nx = ty;
			ny = tx;
		    }
		    else if (between(0,ty,env.height)) {
			sx = env.width;
			sy = ty;
			nx = env.width - (nx - env.width);
		    }
		    else {
			sy = env.height;
			sx = tx;
			ny = env.height - (ny - env.height);
		    }
		}
		else if ((sx > nx) && (sy < ny)) { // second quadrant
		    ty = intersect_x_at(sx,sy,nx,ny,0.0);
		    tx = intersect_y_at(sx,sy,nx,ny,env.height);
		    if ((tx == 0.0) && (ty == env.height)) {
			//intersects at the corner (0,100)
			tx = ny - env.height;
			ty = env.height + nx;
			sx = 0.0;
			sy = env.height;
			nx = ty;
			ny = tx;
		    }
		    else if (between(0.0,ty,env.height)) {
			sx = 0.0;
			sy = ty;
			nx = 0.0 - nx;
		    }
		    else {
			sy = env.height;
			sx = tx;
			ny = env.height - (ny - env.height);
		    }
		}
		else if ((sx > nx) && (sy > ny)) { // third quadrant
		    ty = intersect_x_at(sx,sy,nx,ny,0.0);
		    tx = intersect_y_at(sx,sy,nx,ny,0.0);
		    if ((tx == 0.0) && (ty == 0.0)) {
			// intersects at the corner (0,0)
			tx = 0.0 - ny;
			ty = 0.0 - nx;
			sx = 0.0;
			sy = 0.0;
			nx = ty;
			ny = tx;
		    }
		    else if (between(0.0,ty,env.height)) {
			sx = 0.0;
			sy = ty;
			nx = 0.0 - nx;
		    }
		    else {
			sy = 0.0;
			sx = tx;
			ny = 0.0 - ny;
		    }
		}
		else { // fourth quadrant
		    ty = intersect_x_at(sx,sy,nx,ny,env.width);
		    tx = intersect_y_at(sx,sy,nx,ny,0.0);
		    if ((tx == env.width) && (ty == 0.0)) {
			// right in the corner (100,0)
			tx = env.width - (0.0 - ny);
			ty = env.height - (env.width - (nx - env.width));
			sx = env.width;
			sy = 0.0;
			nx = tx;
			ny = ty;
		    }
		    else if (between(0.0,ty,env.height)) {
			sx = env.width;
			sy = ty;
			nx = env.width - (nx - env.width);
		    }
		    else {
			sy = 0.0;
			sx = tx;
			ny = 0.0 - ny;
		    }

		}

		if (env.Debug > 2) {
		    System.out.println("In while loop ...");
		    System.out.println("Start sx " + sx + " sy " + sy);
		    System.out.println("End nx " + nx + " ny " + ny);
		}
	    }
	    if (env.Debug > 0) {
		System.out.println ("Mobile Object " + name + " new location: (" + x_coord + "," + y_coord + ")");
		// System.out.println ("  Local Radian value: " + radian);
	    }
	    x_coord = nx;
	    y_coord = ny;

	    // Only update the direction if location has changed
	    if ( (nx - sx) != 0 || (ny - sy) != 0 )
		mdirection = new_direction(nx - sx, ny - sy);

	    // Only redraw graphics when needed (BCH)
	    if ((ox != env.scale_x_coord ( x_coord, (double) lastw)) ||
		(oy != env.scale_y_coord ( y_coord, (double) lasth)))
		env.requestRedraw();
	}
    }

    public double new_direction (double x, double y) {
	    double r = 0;
	    double h = Math.sqrt((x * x) + (y * y));

	    if ((x > 0) && (y > 0)) {
		r = Math.asin(y / h);
	    }
	    else if ((x < 0) && (y > 0)) {
		r = Math.PI - Math.asin(y / h);
	    }
	    else if ((x < 0) && (y < 0)) {
		r = Math.PI + Math.abs( Math.asin(y / h) );
	    }
	    else if ((x > 0) && (y < 0)) {
		r = (2 * Math.PI) - Math.asin(Math.abs(y) / h);
	    }
	    else if ((x > 0) && (y == 0)) {
		r = 0.0;
	    }
	    else if ((x < 0) && (y == 0)) {
		r = Math.PI;
	    }
	    else if ((x == 0) && (y > 0)) {
		r = Math.PI / 2;
	    }
	    else if ((x == 0) && (y < 0)) {
		r = (3 * Math.PI) / 2;
	    }

	    return(((r * 180.0) / Math.PI));
	}

    public double get_x_coord () {
	return x_coord;
    }

    public double get_y_coord () {
	return y_coord;
    }


}
